<?LassoScript

/////////////////////////////////////
//
//	PROCS
//	Used to filter response data from a SOAP server.
//	Many different procs can be given, in an array when a SOAP operation is bound to a tag name.
//	When the SOAP server response data is received, the SOAP:Body XML node is extracted and passed
//	to each proc in the list.
//
//	Procs accept 1 parameter* and should return one parameter. Their purpose is to transform the input data in some way,
//	and return that transformation. The next proc in the list will be ->invoked with that resulting data.
//	In this way, each proc will have a chance to handle the data before passing it on to the next.
//
//	Procs can do activities such as (but not limited to):
//		* Evaluating an XPath expression
//		* Executing a regular expressions
//		* Executing XSLT
//		* Returning specific items from lists
//		* Converting XML data into Lasso type instances dynamically, utilizing XMLSchema
//		* ...
//
//	Custom procs which are either (custom or LCAPI) tag types, or custom types, 
//	can be easily authored and introduced into the processing system.
//
//
//	* Procs may be passed further information in the future
//
/////////////////////////////////////

define_type:'proc_map_entry';
	local:'name'='';
	local:'proc'=null;
	local:'description'='';
	local:'isinstance'=boolean;
	local:'hasparams'=boolean;
		
	define_tag:'onCreate', -required='name', -required='proc', -type='tag', -required='isinstance', -required='hasparams';
		self->'name' = #name;
		self->'proc' = #proc;
		self->'description' = #proc->description;
		self->'isinstance' = #isinstance; // is it a type instance that should be created for each use or is it a simple tag which should be invoked for each use
		self->'hasparams' = #hasparams;
	/define_tag;
/define_type;

define_type:'_proc_withParam', -description="Base for a proc which stores a parameter value along with it. "
											"Implement 'invoke' to add custom behavior.";
	local:'param'=null;	
	define_tag:'onCreate';
		self->param = params->first;
	/define_tag;
/define_type;

define_type:'proc_extract', '_proc_withParam', -description="Executes XML->extract on the parameter. "
															"Create the type with the XPath expression to evaluate.";
	define_tag:'invoke', -required='xml', -type='xml';
		return:#xml->extract(self->param);
	/define_tag;
/define_type;

define_type:'proc_extractOne', '_proc_withParam', -description="Executes XML->extractOne on the parameter. "
															"Create the type with the XPath expression to evaluate.";
	define_tag:'invoke', -required='xml', -type='xml';
		return:#xml->extractOne(self->param);
	/define_tag;
/define_type;

define_type:'proc_get', '_proc_withParam', -description="Executes ->get on the parameter. "
														"Create the type with the position to 'get'.";
//	define_tag:'invoke', -required='gettablbasket
//	e';
	define_tag:'invoke', -required='gettable';
		return:#gettable->get(self->param);
	/define_tag;
/define_type;

define_tag:'proc_xml', -required='something', -description=	"Converts the parameter to the XML type.";
	return:xml(#something);
/define_tag;

define_tag:'proc_null', -description="Passes the input type back unaltered.";
	return:params->first;
/define_tag;

// xslt
// regexp
// ->first
// ->second
// proc_choose:expression_proc, true_proc, false_proc;

define_tag:'proc_convert', -required='xml', -type='xml', -description=	"Converts the XML parameter to a Lasso type instance. "
																		"The will be one single element. "
																		"This is the default proc that will be applied if no other procs are given.";
	return:xsd_generate(#xml);
/define_tag;

define_tag:'proc_convertOne', -required='xml', -type='xml', -description=	"Converts the first child of the XML parameter to a Lasso type instance. "
																		"The will be one single element.";
	return:xsd_generate(#xml->extractOne('*'));
/define_tag;

define_tag:'proc_convertBody', -required='xml', -type='xml', -description=	"Converts the XML parameter to an ARRAY of Lasso type instances. "
																		"The will (generally) be one element in the array for each SOAP:Body level XML node. "
																		"This is the default proc that will be applied if no other procs are given.";
	local:'ret'=array;
	iterate:#xml->extract('*'), local:'x';
		#ret->insert(xsd_generate(#x));
	/iterate;
	return:#ret;
/define_tag;

define_type:'_soap_info';
	local:'operation'='';
	local:'action'='';
	local:'location'='';
	local:'targetNamespace'=''
	local:'input'=array;
	local:'output'=array;
	local:'use'='';
	local:'defaults'=null;
	local:'procs'=null;
	
	define_tag:'onCreate', 	-required='opname', 
							-required='soapaction', 
							-required='location',
							-required='targetNamespace',
							-required='input', -type='array', 
							-required='output', -type='array',
							-required='use',
							-optional='defaults',
							-optional='procs';
		self->operation = #opname;
		self->action = #soapaction;
		self->location = #location;
		self->targetNamespace = #targetNamespace;
		self->input->merge(#input);
		self->output->merge(#output);
		self->use = #use;
		self->defaults = local:'defaults';
		if: local('procs') == null;
			self->procs = array(\proc_convertBody); // default
		else if: !#procs->isa('array');
			self->procs = array(#procs);
		else;
			self->procs = #procs;
		/if;
	/define_tag;
	
	define_tag:'reset';
		self->operation = '';
		self->action = '';
		self->location = '';
		self->targetNamespace = '';
		self->input = array;
		self->output = array;
		self->use = '';
		self->defaults = map;
		self->procs = array;
	/define_tag;
/define_type;

define_prototype:'soap_info', _soap_info('', '', '', '', array, array, '', map, array);

define_tag:'_soap_stub', -required='_soap_local_info', -type='soap_info';
	// do stuff
	// should have enough information to create the soap 
	// request and process and return the soap response
	// be careful naming local variables so that they won't clash with explicitly named soap parameters
//	return:#_soap_local_info->procs;
	// make a unique list of namespaces for each type used
	// special case for xmlschema types so they always have xsd namespace
	local:'nsmap'=map; // keyed by uri->prefix
	#nsmap->insert($XSD = 'xsd');
	iterate:#_soap_local_info->input, local:'_soap_local_curr';
		local:'_soap_local_uri' = #_soap_local_curr->second->namespace;
		if: #_soap_local_uri->length && #_soap_local_uri != $XSD && #nsmap !>> #_soap_local_uri;
			#nsmap->insert(#_soap_local_uri = ('ns' + loop_count) );
		/if;
	/iterate;
	
	local:'_soap_local_outgoing' = '<?xml version="1.0" encoding="UTF-8"?>\r\n\
			<SOAP-ENV:Envelope xmlns:SOAP-ENV="http://schemas.xmlsoap.org/soap/envelope/" xmlns:xsi="http://www.w3.org/1999/XMLSchema-instance"';
	// add the automatically generated namespaces
	iterate:#nsmap, #_soap_local_curr;
		#_soap_local_outgoing += ' xmlns:' #_soap_local_curr->second '="' #_soap_local_curr->first '"';
	/iterate;
		
	#_soap_local_outgoing += '>\r\n<SOAP-ENV:Body>\r\n';

	#_soap_local_outgoing += '<targns:' #_soap_local_info->operation ' xmlns:targns="' #_soap_local_info->targetNamespace '" SOAP-ENV:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">\r\n';
	
	local:'idx'=1; // set to one because the first param we receive is always the soap_info
	
	// loop here through all the 'parts' of the soap input message
	// each one is either a named parameter, positional parameter, or a pre-established default
	iterate:#_soap_local_info->input, #_soap_local_curr;
		local:'_soap_local_val'=null; // null can be a valid value, so we have to mark it with...
		// this
		local:'found' = false;
		
		if:local_defined(#_soap_local_curr->first); // named explicitly
			#_soap_local_val = string(local(#_soap_local_curr->first));
			#found = true;
		else: (#_soap_local_info->defaults >> #_soap_local_curr->first); // a default was provided
			#_soap_local_val = #_soap_local_info->defaults->find(#_soap_local_curr->first);
			#found = true;
		else;
			// find the next unnamed param
			while: ++#idx <= params->size;
				if:!params->get(#idx)->isa('pair'); // valid. pairs are indicative of name/value. 
													// we can only consider value since we seek un-named params
					#_soap_local_val = string(params->get(#idx));
					#found = true;
					loop_abort;
				/if;				
			/while;			
		/if;		
		
		#_soap_local_outgoing += '<' #_soap_local_curr->first ' xsi:type="';
		#_soap_local_outgoing += (#nsmap->find(#_soap_local_curr->second->namespace) + ':' + #_soap_local_curr->second->name) '"';
		if:#_soap_local_val->type == 'null';
			#_soap_local_outgoing += ' />\r\n';
		else;
			#_soap_local_outgoing += '>' encode_xml(#_soap_local_val) '</' #_soap_local_curr->first '>\r\n';	
		/if;
	/iterate;
	
	#_soap_local_outgoing += '</targns:' #_soap_local_info->operation '>\r\n\
	</SOAP-ENV:Body></SOAP-ENV:Envelope>';
	
//	return:#_soap_local_outgoing;
	
	local:'response' = xml:include_url(#_soap_local_info->location, 
			-SendMIMEHeaders=array('Content-Type'='text/xml; charset=UTF-8', 'SOAPAction'='"' + #_soap_local_info->action + '"',
				'Connection'='Close'),
			-POSTParams=#_soap_local_outgoing);
//	return:#response;

	local:'fault' = #response->extractOne("/*[local-name()='Envelope']/*[local-name()='Body']/*[local-name()='Fault']"); /* */
	if:#fault != null;
		fail:-1, #fault->extractOne("*[local-name()='faultcode']")->contents ': ' #fault->extractOne("*[local-name()='faultstring']")->contents;
	/if;
	
	local:'body' = #response->extractOne("/*[local-name()='Envelope']/*[local-name()='Body']"); /* */
	iterate:#_soap_local_info->procs, local:'proc';
		#body = #proc->invoke(#body);
	/iterate;
	return:#body;
/define_tag;

define_tag:'soap_convertPartsToPairs', -required='xml', -type='xml', -required='parts', -type='array';
	local:'ret'=array;
	iterate:#parts, local:'i';
		local:'type' = #i->extract('normalize-space(@type)');
		if:#type == null;
			#type = #i->extract('normalize-space(@element)');
		/if;
		#ret->insert(pair(#i->extract('normalize-space(@name)') = xsd_ref(xsd_fullyQualifyName(#xml, '', #type))));
	/iterate;
	return:#ret;
/define_tag;

define_tag:'soap_definetag', 
						-required='localTagName', -type='string', 
						-required='wsdl', -type='xml',
						-required='operationName', -type='string',
						-optional='defaults',
						-optional='procs',
							-description='Maps a SOAP Client operation to a Lasso tag name. Calling the tag will invoke the SOAP operation and return the response.';
    local:'_start' = _date_msec;
    xsd_loadSchema(#wsdl);
    local:'op' = @wsdl_getOperation(#wsdl, #operationName);
    fail_if: #op == null, -1, 'Could not find the requested operation.';
    
    local:'msgtypes' = @wsdl_getOpMessageTypes(#op);
    fail_if:#msgtypes == null, -1, 'Could not get message types.';
    
    local:'info' = soap_info(#operationName, 
    					wsdl_getOpSoapAction(#op), 
    					wsdl_getOpLocation(#wsdl, #operationName), 
    					wsdl_getTargetNamespace(#wsdl),
    					soap_convertPartsToPairs(#wsdl, wsdl_getMessageParts(#wsdl, xsd_ref(#msgtypes->first)->name)), 
    					soap_convertPartsToPairs(#wsdl, wsdl_getMessageParts(#wsdl, xsd_ref(#msgtypes->second)->name)),
    					wsdl_getOpBodyUse(#op),
    					local('defaults'),
    					local('procs') );
        
    $__TAGS__->insert(#localTagName = @pair(\_soap_stub = @#info) );
    log_detail:'soap_definetag: ' + (_date_msec - #_start) + 'ms';
/define_tag;

if:global('soap_proc_list')->size == 0;
	$soap_proc_list = map(
							"nothing"=proc_map_entry('proc_null', \proc_null, false, false),
							"extract"=proc_map_entry('proc_extract', \proc_extract, true, true),
							"extractOne"=proc_map_entry('proc_extractOne', \proc_extractOne, true, true),
							"get"=proc_map_entry('proc_get', \proc_get, true, true),
							"xml"=proc_map_entry('proc_xml', \proc_get, false, false),
							"convert"=proc_map_entry('proc_convert', \proc_convert, false, false),
							"convertOne"=proc_map_entry('proc_convertOne', \proc_convertOne, false, false),
							"convertBody"=proc_map_entry('proc_convertBody', \proc_convertBody, false, false)
							);
/if;

?>